package etsiqkdnclient

import (
	"net"
	"os"
	"time"

	log "github.com/sirupsen/logrus"

	"code.fbi.h-da.de/danet/proto-kms/kms"
	"github.com/google/uuid"
)

// Example of a valid config.yaml
//
// Name: kms02
// InterComAddr: 172.20.10.15:50910
// GRPCAddr: 172.20.10.15:50900
// Peers:
//     - PeerInterComAddr: 172.20.10.14:50910
//       Sync: false
//       QuantumModule:
//         UDPAddr: 172.20.10.15:50915
//         PeerUDPAddr: 172.20.10.14:50915
//         GenerateKeys: false
//     - PeerInterComAddr: 172.20.10.16:50910
//       Sync: true
//       QuantumModule:
//         UDPAddr: 172.20.10.15:50916
//         PeerUDPAddr: 172.20.10.16:50916
//         GenerateKeys: true
//     - PeerInterComAddr: 172.20.10.17:50910
//       Sync: true
//       QuantumModule:
//         UDPAddr: 172.20.10.15:50917
//         PeerUDPAddr: 172.20.10.17:50917
//         GenerateKeys: true

type Config struct {
	Name         string `yaml:"Name"`
	InterComAddr string `yaml:"InterComAddr"`
	GRPCAddr     string `yaml:"GRPCAddr"`
	Peers        []struct {
		PeerInterComAddr string `yaml:"PeerInterComAddr"`
		Sync             bool   `yaml:"Sync"`
		QuantumModule    struct {
			UDPAddr      string `yaml:"UDPAddr"`
			PeerUDPAddr  string `yaml:"PeerUDPAddr"`
			GenerateKeys bool   `yaml:"GenerateKeys"`
		} `yaml:"QuantumModule"`
	} `yaml:"Peers"`
}

type EtsiQkdClient interface {
	GetQkdnVersion() *string
	GetQkdnUui() *string
	GetQkdnLocationId() *string
	GetEKMS() *kms.EKMS
	SubscribeToEtsiQkdClient() chan string
}

// This provides some static information
type etsiQkdnInfo struct {
	// Information used to fill the ETSI GS QKD 15 yang model
	qkdnVersion          string // qkdn_version
	qkdnUuid             string
	qkdnLocationId       string
	kms                  *kms.EKMS
	KmsPeerUpdateChannel chan string // used to get updates from KmsPeer part
}

// A QkdnClient for the emulated KMS
// Creates two mockup QKD Interfaces for testing
func NewQkdnClient(bootInfo *Config) (myInfo *etsiQkdnInfo) {

	myInfo = new(etsiQkdnInfo)
	myInfo.qkdnVersion = "danet-emulated-kms;0.1.0"
	myInfo.qkdnUuid = uuid.NewString()
	myInfo.qkdnLocationId = "toujours et partout"
	myInfo.KmsPeerUpdateChannel = make(chan string)

	log.Printf("%s in regular operations mode", bootInfo.Name)
	myInfo.kms = emulatedKMS(bootInfo, myInfo.KmsPeerUpdateChannel)

	return myInfo
}

// TODO: return an error
func emulatedKMS(config *Config, peerChannel chan string) *kms.EKMS {
	// Attach to eKMS
	emuKMS := kms.NewEKMS(config.Name, uuid.New(), os.Stdout, log.TraceLevel, false, config.InterComAddr)
	emuKMS.AddExternalNotifierKMSPeer(peerChannel)

	for _, peer := range config.Peers {
		// Fire up Quantum LinK
		myQL := emuKMS.AddQuantumElement(peer.QuantumModule.UDPAddr, peer.QuantumModule.GenerateKeys, os.Stdout, log.TraceLevel, false)

		udpQL2Addr, err := net.ResolveUDPAddr("udp", peer.QuantumModule.PeerUDPAddr)
		if err != nil {
			log.Fatalf("QuantumlayerEmuPRNG UDP failure: %s", err)
			return nil
		}

		myQL.QuantumElementLink.AddPeer(udpQL2Addr)

		p, err := emuKMS.AddPeer(peer.PeerInterComAddr, myQL)
		if err != nil {
			log.Fatalf("Failed to create a peer: %s", err)
			return nil
		}

		if peer.Sync {
			go func() {
				time.Sleep(time.Second * 32)
				if err := p.SyncBulkKeys(); err != nil {
					log.Println("SYNC ERROR: ", err)
				}
			}()
		}
	}

	// Start the SDN/management and key retrieval interface
	go kms.StartETSI(config.GRPCAddr, emuKMS)

	// TODO/XXX catch errors!
	go emuKMS.GlobalKeyHandler(7 * time.Second)

	return emuKMS
}

func (qkdnInfo *etsiQkdnInfo) GetQkdnVersion() *string {
	return &qkdnInfo.qkdnVersion
}

func (qkdnInfo *etsiQkdnInfo) GetQkdnUui() *string {
	return &qkdnInfo.qkdnUuid
}

func (qkdnInfo *etsiQkdnInfo) GetQkdnLocationId() *string {
	return &qkdnInfo.qkdnLocationId
}

func (qkdnInfo *etsiQkdnInfo) GetEKMS() *kms.EKMS {
	return qkdnInfo.kms
}

func (qkdnInfo *etsiQkdnInfo) SubscribeToEtsiQkdClient() chan string {
	return qkdnInfo.KmsPeerUpdateChannel
}
