package etsi

import (
	"fmt"

	"code.fbi.h-da.de/danet/ekms/etsiqkdnclient"
	gnmitargetygot "code.fbi.h-da.de/danet/ekms/model"
	"code.fbi.h-da.de/danet/proto-kms/kms"
	"github.com/google/uuid"
	"github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/ygot/ygot"
	log "github.com/sirupsen/logrus"
)

// ETSI15Handler is the implementation of a gnmitarget.PathHandler.
type ETSI15Handler struct {
	name       string
	paths      map[string]struct{}
	qkdnClient etsiqkdnclient.EtsiQkdClient
}

func NewETSI15Handler(client etsiqkdnclient.EtsiQkdClient) *ETSI15Handler {
	return &ETSI15Handler{
		name: "etsi15-handler",
		paths: map[string]struct{}{
			"/qkdn": struct{}{},
		},
		qkdnClient: client,
	}
}

func (yh *ETSI15Handler) Name() string {
	return yh.name
}

func (yh *ETSI15Handler) Paths() map[string]struct{} {
	return yh.paths
}

func (yh *ETSI15Handler) Init(c ygot.ValidatedGoStruct) error {
	config, ok := c.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return fmt.Errorf("failed type assertion for newConfig %T", (*gnmitargetygot.Gnmitarget)(nil))
	}

	//Prepare the KMS to send KMSPeer notifications
	yh.qkdnClient.GetEKMS().AddExternalNotifierKMSPeer(yh.qkdnClient.SubscribeToEtsiQkdClient())

	// Create ygot structs for the ETSI GS QKD 15 models
	confQkdNode := config.GetOrCreateQkdNode()

	// Fill out the known fields in the ygot structs
	err := updateOrCreateQkdNode(confQkdNode, yh.qkdnClient)
	if err != nil {
		return err
	}

	// Start the go routine that takes care of any update from the kms
	go func() {
		kmsChan := yh.qkdnClient.SubscribeToEtsiQkdClient()
		for {
			select {
			case <-kmsChan:
				log.Println("Update for KMSPeer.")
				if err := updateOrCreateQkdNode(confQkdNode, yh.qkdnClient); err != nil {
					fmt.Println("Error within qkdn update goroutine.")
				}
			}
		}
	}()

	return nil
}

func (yh *ETSI15Handler) Update(c ygot.ValidatedGoStruct, jobs []*gnmi.Update) error {
	fmt.Println("Update request received for ", yh.name)
	config, ok := c.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return fmt.Errorf("failed type assertion for newConfig %T", (*gnmitargetygot.Gnmitarget)(nil))
	}

	qkdnInfo := config.GetQkdNode()

	if qkdnInfo != nil {
		// check for changes in the QKD interfaces
		if q := qkdnInfo.GetOrCreateQkdInterfaces(); q != nil {
			for _, qkdnIface := range q.QkdInterface {
				qle, exists := yh.qkdnClient.GetEKMS().QuantumElements[*qkdnIface.QkdiId]
				if !exists {
					log.Printf("Could not find QuantumElement with QKD-ID: %d\n", *qkdnIface.QkdiId)
					return fmt.Errorf("Error: Could not find QuantumElement with QKD-ID: %d\n", *qkdnIface.QkdiId)
				}
				if qkdnIface.QkdiStatus == gnmitargetygot.EtsiQkdSdnNode_IfaceStatusTypes_DISABLED {
					qle.QuantumElementLink.PowerOff()
				} else if qkdnIface.QkdiStatus == gnmitargetygot.EtsiQkdSdnNode_IfaceStatusTypes_ENABLED {
					qle.QuantumElementLink.PowerOn(true)
				}
			}
		}
		if links := qkdnInfo.GetOrCreateQkdLinks(); links != nil {
			theEKMS := yh.qkdnClient.GetEKMS()
			for _, qkdnLink := range links.QkdLink {
				peerUuid, uuidErr := uuid.Parse(*qkdnLink.QkdlId)
				if uuidErr != nil {
					return fmt.Errorf("Couldn't parse uuid %s of request", *qkdnLink.QkdlId)
				}
				// Lookup peer and skip if there
				// TODO needs handling for existing peers
				if thePeer := theEKMS.FindPeerUuid(peerUuid); thePeer == nil {
					// should be changed!!!!
					if qkdnLink.QkdlLocal.QkdiId != nil && qkdnLink.QkdlRemote.QkdiRemoteSocket != nil && qkdnLink.QkdlStatus == gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_ACTIVE {
						if elementID, _ := theEKMS.QuantumElements[*qkdnLink.QkdlLocal.QkdiId]; elementID != nil {
							theEKMS.AddPeer(*qkdnLink.QkdlRemote.QkdiRemoteSocket, elementID)
						}
					}
				}
				if qkdnLink.QkdlLocal.QkdiId != nil && qkdnLink.QkdlRemote.QkdiRemoteSocket != nil && qkdnLink.QkdlStatus == gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_PASSIVE {
					theEKMS.RemovePeer(*qkdnLink.QkdlRemote.QkdiRemoteSocket)
				}
			}
		}
	}
	return nil
}

func updateOrCreateQkdNode(confQkdNode *gnmitargetygot.EtsiQkdSdnNode_QkdNode, qkdInfo etsiqkdnclient.EtsiQkdClient) error {
	// set current version of the QDKN module
	confQkdNode.QkdnVersion = qkdInfo.GetQkdnVersion()
	confQkdNode.QkdnId = qkdInfo.GetQkdnUui()
	confQkdNode.QkdnLocationId = qkdInfo.GetQkdnLocationId()
	/*
		confQkdNode.QkdnCapabilities.ApplicationStatsSupport
		confQkdNode.QkdnCapabilities.LinkStatsSupport
		confQkdNode.QkdnCapabilities.KeyRelayModeEnable
	*/

	// Fill QKD Interfaces with the relevant information
	// Interfaces structure
	confQkdNode.QkdInterfaces = confQkdNode.GetOrCreateQkdInterfaces()
	QkdiModel := "danet-emulated-qle-prng"

	// range through existing QKD interfaces out of proto-kms
	myKMS := qkdInfo.GetEKMS()
	for _, element := range myKMS.QuantumElements {
		log.Printf("Element with ID %d\n", element.QlID)

		//Check if element exists in ygot struct
		qkdIface, nodeErr := confQkdNode.QkdInterfaces.NewQkdInterface(element.QlID)
		if nodeErr != nil {
			qkdIface = confQkdNode.QkdInterfaces.GetQkdInterface(element.QlID)
		} else {
			confQkdNode.QkdInterfaces.AppendQkdInterface(qkdIface)
		}

		qkdIface.QkdiId = &element.QlID
		if poweredOn, _ := element.QuantumElementLink.GetStatus(); poweredOn == true {
			qkdIface.QkdiStatus = gnmitargetygot.EtsiQkdSdnNode_IfaceStatusTypes_ENABLED
		} else {
			qkdIface.QkdiStatus = gnmitargetygot.EtsiQkdSdnNode_IfaceStatusTypes_DISABLED
		}
		qkdIface.QkdiModel = &QkdiModel
		//qkdIface.QkdiAttPoint.Device =
		element.QuantumElementLink.GetLocalQLPort()
	}
	confQkdNode.QkdLinks = confQkdNode.GetOrCreateQkdLinks()

	// Check for changes in the QKD key association links
	if q := confQkdNode.GetOrCreateQkdLinks(); q != nil {
		for _, kmsLink := range qkdInfo.GetEKMS().KmsPeers {
			// Just write the current status of this kms peer
			qkdnLink := confQkdNode.QkdLinks.GetOrCreateQkdLink(kmsLink.GetKmsPeerId().String())
			switch kmsLink.GetKmsPeerStatus() {
			case kms.KmsPeerDown:
				qkdnLink.QkdlStatus = gnmitargetygot.E_EtsiQkdSdnNode_LinkStatusTypes(gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_OFF)
			case kms.KmsPeerUp:
				qkdnLink.QkdlStatus = gnmitargetygot.E_EtsiQkdSdnNode_LinkStatusTypes(gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_ACTIVE)
			case kms.KmsPeerPending:
				qkdnLink.QkdlStatus = gnmitargetygot.E_EtsiQkdSdnNode_LinkStatusTypes(gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_PENDING)
			default:
				qkdnLink.QkdlStatus = gnmitargetygot.E_EtsiQkdSdnNode_LinkStatusTypes(gnmitargetygot.EtsiQkdSdnNode_LinkStatusTypes_UNSET)
			}
			peerQkdiId := kmsLink.GetKmsPeerQkdiId()
			qkdLocal := qkdnLink.GetOrCreateQkdlLocal()
			qkdLocal.QkdiId = &peerQkdiId

			qkdnLink.QkdlLocal = qkdLocal
		}
	}

	confQkdNode.QkdApplications = confQkdNode.GetOrCreateQkdApplications()

	return nil
}
